#!/usr/bin/python3
# Dogtail test script for addtrans.

import os
import pipes
import tempfile

from dogtail import config
from dogtail import tree
from dogtail.procedural import type
from dogtail.rawinput import keyCombo, pressKey
from dogtail.utils import run


os.environ['LANG'] = "en_US.UTF-8"
os.environ['PYTHONPATH'] = os.path.join(
    os.path.dirname(__file__),
    os.path.pardir,
    os.path.pardir,
    'src'
)
os.environ['PATH'] =  os.path.join(
    os.path.dirname(__file__),
    os.path.pardir,
    os.path.pardir,
    'bin'
) + os.path.pathsep + os.environ['PATH']

config.config.typingDelay = 0.025

t = tempfile.NamedTemporaryFile()
t.write("""
2015-10-05 * beer
    Assets:Cash                       -30 CHF
    Expenses:Drinking                  30 CHF
""")
t.flush()
t.seek(0, 0)

run('addtrans --file %s' % pipes.quote(t.name))
addtrans = tree.root.application('addtrans')
mainwin = addtrans.window('Add transaction')

try:
    type("wine")
    pressKey("Tab")
    type("30")
    pressKey("Tab")
    type("Expenses:Drinking")
    pressKey("Tab")
    pressKey("Tab")
    type("Assets:Cash")
    pressKey("Tab")

    expected = """
2016-11-09 wine
    Expenses:Drinking    30 CHF
    Assets:Cash
"""

    actual = mainwin.child(name='Transaction preview').children[0].text
    assert (
        actual == expected
    ), (
        "Transaction preview did not contain 30 CHF as expected.\n"
        "Expected: %r\n"
        "Actual: %r" % (expected, actual)
    )
finally:
    keyCombo("<Alt>c")

#def recurse(child, level=0):
    #try:
        #print "   " * level, child.getAbsoluteSearchPath()
        #print "   " * level, child.text
    #except UnicodeDecodeError:
        #print "   " * level, "[undecodable path]"
    #for c in child.children:
        #recurse(c, level+1)

#recurse(mainwin)



#!/usr/bin/python3

import argparse
import datetime
import logging
import traceback

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

import ledgerhelpers as common
from ledgerhelpers import gui
from ledgerhelpers import journal
from ledgerhelpers.programs import common as common_programs
import ledgerhelpers.editabletransactionview as ed


ASYNC_LOAD_MESSAGE = "Loading completion data from your ledger..."
ASYNC_LOADING_ACCOUNTS_MESSAGE = (
    "Loading account and commodity data from your ledger..."
)


class AddTransWindow(Gtk.Window, gui.EscapeHandlingMixin):

    def __init__(self):
        Gtk.Window.__init__(self, title="Add transaction")
        self.set_border_width(12)

        grid = Gtk.Grid()
        grid.set_column_spacing(8)
        grid.set_row_spacing(12)
        self.add(grid)

        row = 0

        self.transholder = ed.EditableTransactionView()
        grid.attach(self.transholder, 0, row, 2, 1)

        row += 1

        self.transaction_view = gui.LedgerTransactionView()
        self.transaction_view.set_vexpand(True)
        self.transaction_view.get_accessible().set_name("Transaction preview")
        grid.attach(self.transaction_view, 0, row, 2, 1)

        row += 1

        self.status = Gtk.Label()
        self.status.set_line_wrap(True)
        self.status.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.status.set_hexpand(True)
        grid.attach(self.status, 0, row, 1, 1)

        button_box = Gtk.ButtonBox()
        button_box.set_layout(Gtk.ButtonBoxStyle.END)
        button_box.set_spacing(12)
        button_box.set_hexpand(False)
        self.close_button = Gtk.Button(stock=Gtk.STOCK_CLOSE)
        button_box.add(self.close_button)
        self.add_button = Gtk.Button(stock=Gtk.STOCK_ADD)
        button_box.add(self.add_button)
        grid.attach(button_box, 1, row, 1, 1)
        self.add_button.set_can_default(True)
        self.add_button.grab_default()


class AddTransApp(AddTransWindow, gui.EscapeHandlingMixin):

    logger = logging.getLogger("addtrans")
    internal_parsing = []

    def __init__(self, journal, preferences):
        AddTransWindow.__init__(self)
        self.journal = journal
        self.preferences = preferences
        self.successfully_loaded_accounts_and_commodities = False

        self.accounts = []
        self.commodities = dict()
        self.internal_parsing = []
        self.payees = []

        self.activate_escape_handling()

        self.close_button.connect("clicked",
                                  lambda _: self.emit('delete-event', None))
        self.add_button.connect("clicked",
                                lambda _: self.process_transaction())
        date = self.preferences.get("last_date", datetime.date.today())
        self.transholder.set_transaction_date(date)
        self.transholder.connect(
            "payee-changed",
            self.payee_changed
        )
        self.transholder.connect(
            "changed",
            self.update_transaction_view
        )

        self.add_button.set_sensitive(False)
        self.transholder.title_grab_focus()
        self.status.set_text(ASYNC_LOAD_MESSAGE)

        self.connect("delete-event", lambda _, _a: self.save_preferences())
        self.reload_completion_data()

    def reload_completion_data(self):
        gui.g_async(
            lambda: self.journal.internal_parsing(),
            lambda payees: self.internal_parsing_loaded(payees),
            self.journal_load_failed,
        )

    def internal_parsing_loaded(self, internal_parsing):
        self.internal_parsing = internal_parsing
        gui.g_async(
            lambda: self.journal.all_payees(),
            lambda payees: self.all_payees_loaded(payees),
            self.journal_load_failed,
        )

    def all_payees_loaded(self, payees):
        self.payees = payees
        self.transholder.set_payees_for_completion(self.payees)
        gui.g_async(
            lambda: self.journal.accounts_and_last_commodity_for_account(),
            lambda r: self.accounts_and_last_commodities_loaded(*r),
            self.journal_load_failed,
        )
        if self.status.get_text() == ASYNC_LOAD_MESSAGE:
            self.status.set_text(ASYNC_LOADING_ACCOUNTS_MESSAGE)

    def accounts_and_last_commodities_loaded(self, accounts, last_commos):
        self.accounts = accounts
        self.commodities = last_commos
        self.transholder.set_accounts_for_completion(self.accounts)
        self.transholder.set_default_commodity_getter(
            self.get_commodity_for_account
        )
        self.successfully_loaded_accounts_and_commodities = True
        if self.status.get_text() == ASYNC_LOADING_ACCOUNTS_MESSAGE:
            self.status.set_text("")

    def journal_load_failed(self, e):
        traceback.print_exception(e)
        gui.FatalError(
            "Add transaction loading failed",
            "An unexpected error took place:\n%s" % e,
        )
        self.emit('delete-event', None)

    def get_commodity_for_account(self, account_name):
        try:
            return self.commodities[account_name]
        except KeyError:
            pass

    def payee_changed(self, emitter=None):
        if emitter.postings_modified() and not emitter.postings_empty():
            return
        text = emitter.get_payee_text()
        self.try_autofill(emitter, text)

    def try_autofill(self, transaction_view, autofill_text):
        ts = journal.transactions_with_payee(
            autofill_text,
            self.internal_parsing,
            case_sensitive=False
        )
        if not ts:
            return
        return self.autofill_transaction_view(transaction_view, ts[-1])

    def autofill_transaction_view(self, transaction_view, transaction):
        transaction_view.replace_postings(transaction.postings)
        transaction_view.set_clearing(transaction.state)

    def update_transaction_view(self, unused_ignored=None):
        self.update_validation()
        k = self.transholder.get_data_for_transaction_record
        title, date, clear, statechar, lines = k()
        self.transaction_view.generate_record(
             title, date, clear, statechar, lines
         )

    def update_validation(self, grab_focus=False):
        try:
            self.transholder.validate(grab_focus=grab_focus)
            self.status.set_text("")
            self.add_button.set_sensitive(True)
            return True
        except common.TransactionInputValidationError as e:
            self.status.set_text(str(e))
            self.add_button.set_sensitive(False)
            return False

    def process_transaction(self):
        if not self.update_validation(True):
            return
        buf = self.transaction_view.get_buffer()
        text = buf.get_text(buf.get_start_iter(), buf.get_end_iter(), True)
        self.journal.add_text_to_file(text)
        self.reset_after_save()

    def reset_after_save(self):
        self.transholder.clear()
        self.transholder.title_grab_focus()
        self.status.set_text("Transaction saved")
        self.reload_completion_data()

    def save_preferences(self):
        if not self.successfully_loaded_accounts_and_commodities:
            return
        self.preferences["default_to_clearing"] = (
            self.transholder.clearing.get_state() !=
            self.transholder.clearing.STATE_UNCLEARED
        )
        if self.transholder.when.get_date() in (datetime.date.today(), None):
            del self.preferences["last_date"]
        else:
            self.preferences["last_date"] = (
                self.transholder.when.get_date()
            )
        self.preferences.persist()


def get_argparser():
    parser = argparse.ArgumentParser(
        'Add new transactions to your Ledger file',
        parents=[common_programs.get_common_argparser()]
    )
    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='activate debugging')
    return parser


def main():
    args = get_argparser().parse_args()
    common.enable_debugging(args.debug)

    GObject.threads_init()

    journal, s = gui.load_journal_and_settings_for_gui(
        ledger_file=args.file,
        price_file=args.pricedb,
    )
    klass = AddTransApp
    win = klass(journal, s)
    win.connect("delete-event", Gtk.main_quit)
    GObject.idle_add(win.show_all)
    Gtk.main()



#!/usr/bin/python3

import subprocess
import tempfile


def three_way_diff(basefilename, leftcontents, rightcontents):
    """Given a file which is assumed to be utf-8, two utf-8 strings,
    left and right, launch a three-way diff.

    Raises: subprocess.CalledProcessError."""
    if isinstance(leftcontents, str):
        leftcontents = leftcontents.encode("utf-8")
    if isinstance(rightcontents, str):
        rightcontents = rightcontents.encode("utf-8")

    prevfile = tempfile.NamedTemporaryFile(prefix=basefilename + ".previous.")
    prevfile.write(leftcontents)
    prevfile.flush()
    newfile = tempfile.NamedTemporaryFile(prefix=basefilename + ".new.")
    newfile.write(rightcontents)
    newfile.flush()
    try:
        subprocess.check_call(
            ('meld', prevfile.name, basefilename, newfile.name)
        )
    finally:
        prevfile.close()
        newfile.close()


def two_way_diff(leftcontents, rightcontents):
    """Given two strings which are assumed to be utf-8, open a
    two-way diff view with them.

    Raises: subprocess.CalledProcessError."""
    if isinstance(leftcontents, str):
        leftcontents = leftcontents.encode("utf-8")
    if isinstance(rightcontents, str):
        rightcontents = rightcontents.encode("utf-8")

    prevfile = tempfile.NamedTemporaryFile(prefix=".base.")
    prevfile.write(leftcontents)
    prevfile.flush()
    newfile = tempfile.NamedTemporaryFile(prefix=".new.")
    newfile.write(rightcontents)
    newfile.flush()
    try:
        subprocess.check_call(
            ('meld', prevfile.name, newfile.name)
        )
    finally:
        prevfile.close()
        newfile.close()



import codecs
import os

import ledgerhelpers as m


if os.getenv('LEDGERHELPERS_TEST_DEBUG'):
    m.enable_debugging(True)


def datapath(filename):
    return os.path.join(os.path.dirname(__file__), "testdata", filename)


def data(filename):
    with codecs.open(datapath(filename), "rb", "utf-8") as f:
        return f.read()



import datetime
import ledgerhelpers.parser as parser
import tests.test_base as base
from unittest import TestCase as T


class TestParser(T):

    def test_simple_transaction(self):
        c = base.data("simple_transaction.dat")
        items = parser.lex_ledger_file_contents(c)
        self.assertEqual(len(items), 3)
        for n, tclass in enumerate([
            parser.TokenWhitespace,
            parser.TokenTransaction,
            parser.TokenWhitespace,
        ]):
            self.assertIsInstance(items[n], tclass)
        transaction = items[1]
        self.assertEqual(transaction.date, datetime.date(2015, 3, 12))
        self.assertEqual(transaction.clearing_date, datetime.date(2015, 3, 15))
        self.assertEqual(transaction.payee, "beer")
        for n, (ac, am) in enumerate([
            ("Accounts:Cash", "-6.00 CHF"),
            ("Expenses:Drinking", "6.00 CHF"),
        ]):
            self.assertEqual(transaction.postings[n].account, ac)
            self.assertEqual(transaction.postings[n].amount, am)

    def test_no_end_value(self):
        c = base.data("no_end_value.dat")
        items = parser.lex_ledger_file_contents(c)
        self.assertEqual(len(items), 5)
        for n, tclass in enumerate([
            parser.TokenWhitespace,
            parser.TokenTransaction,
            parser.TokenWhitespace,
            parser.TokenTransaction,
            parser.TokenWhitespace,
        ]):
            self.assertIsInstance(items[n], tclass)
        for transaction in (items[1], items[3]):
            self.assertEqual(transaction.payee, "beer")
            for n, (ac, am) in enumerate([
                ("Accounts:Cash", "-6.00 CHF"),
                ("Expenses:Drinking", ""),
            ]):
                self.assertEqual(transaction.postings[n].account, ac)
                self.assertEqual(transaction.postings[n].amount, am)

    def test_with_comments(self):
        c = base.data("with_comments.dat")
        items = parser.lex_ledger_file_contents(c)
        self.assertEqual(len(items), 3)
        for n, tclass in enumerate([
            parser.TokenWhitespace,
            parser.TokenTransaction,
            parser.TokenWhitespace,
        ]):
            self.assertIsInstance(items[n], tclass)
        transaction = items[1]
        self.assertEqual(transaction.date, datetime.date(2011, 12, 25))
        self.assertEqual(transaction.clearing_date, datetime.date(2011, 12, 25))
        self.assertEqual(transaction.payee, "a gift!")
        self.assertEqual(transaction.state, parser.STATE_CLEARED)
        for n, (ac, am) in enumerate([
            ("Assets:Metals", "1 \"silver coin\"    @ $55"),
            ("Income:Gifts", "$        -55"),
        ]):
            self.assertEqual(transaction.postings[n].account, ac)
            self.assertEqual(transaction.postings[n].amount, am)

    def test_my_data_file(self):
        try:
            c = base.data("/home/user/.ledger")
        except IOError:
            return
        items = parser.lex_ledger_file_contents(c)



#!/usr/bin/python3
# coding: utf-8

import gi; gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from ledgerhelpers import parser


class TransactionStateButton(Gtk.Button):

    STATE_CLEARED = parser.STATE_CLEARED
    STATE_UNCLEARED = parser.STATE_UNCLEARED
    STATE_PENDING = parser.STATE_PENDING

    def __init__(self):
        Gtk.Button.__init__(self)
        self.label = Gtk.Label()
        self.add(self.label)
        self.state = "uninitialized"
        self.connect("clicked", lambda _: self._rotate_state())
        self.get_style_context().add_class("circular")
        self._rotate_state()

    def _rotate_state(self):
        if self.state == self.STATE_UNCLEARED:
            self.state = self.STATE_CLEARED
        elif self.state == self.STATE_CLEARED:
            self.state = self.STATE_PENDING
        else:
            self.state = self.STATE_UNCLEARED
        self._reflect_state()

    def _reflect_state(self):
        addtext = "\n\nToggle this to change the transaction state."
        if self.state == self.STATE_UNCLEARED:
            self.label.set_markup("∅")
            self.set_tooltip_text(
                "This transaction is uncleared." + addtext
            )
        elif self.state == self.STATE_CLEARED:
            self.label.set_markup("✻")
            self.set_tooltip_text(
                "This transaction is cleared." + addtext
            )
        else:
            self.label.set_markup("!")
            self.set_tooltip_text(
                "This transaction is pending." + addtext
            )

    def get_state(self):
        return self.state

    def get_state_char(self):
        if self.state == parser.STATE_CLEARED:
            clearing_state = parser.CHAR_CLEARED
        elif self.state == parser.STATE_UNCLEARED:
            clearing_state = ""
        elif self.state == parser.STATE_PENDING:
            clearing_state = parser.CHAR_PENDING
        else:
            assert 0, "not reached"
        return clearing_state

    def set_state(self, state):
        assert state in (self.STATE_CLEARED,
                         self.STATE_PENDING,
                         self.STATE_UNCLEARED)
        self.state = state
        self._reflect_state()



#!/usr/bin/python3

import datetime
import fnmatch
import ledger
import re
import subprocess
import sys
import ledgerhelpers.legacy as common
from ledgerhelpers import gui


class Lot(object):

    def __init__(self, number, date, amount, acct):
        self.number = number
        self.date = date
        self.amount = amount
        quantity = ledger.Amount(amount.strip_annotations())
        self.price = self.amount.price() / quantity
        self.account = acct

    def __str__(self):
        return (
            "<Lot %s of %s at price %s in acct "
            "%s with date %s>"
        ) % (
            self.number, self.amount, self.price, self.account,
            self.date
        )


class NotEnough(Exception):
    pass


class Lots(object):

    def __init__(self):
        self.lots = []
        self.unfinished = []

    def __getitem__(self, idx):
        return list(self)[idx]

    def __iter__(self):
        return iter(sorted(
            self.lots,
            key=lambda l: "%s" % (l.date,)
        ))

    def parse_ledger_bal(self, text):
        """Demands '--balance-format=++ %(account)\n%(amount)\n' format.
        Demands '--date-format=%Y-%m-%d' date format."""
        lines = [x.strip() for x in text.splitlines() if x.strip()]
        account = None
        for line in lines:
            if line.startswith("++ "):
                account = line[3:]
            else:
                amount = ledger.Amount(line)
                date = re.findall(r'\[\d\d\d\d-\d\d-\d\d]', line)
                assert len(date) < 2
                if date:
                    date = common.parse_date(date[0][1:-1])
                else:
                    date = None
                try:
                    lot = Lot(self.nextnum(),
                            date,
                            amount,
                            account)
                    self.lots.append(lot)
                except TypeError:
                    # At this point, we know the commodity does not have a price.
                    # So we ignore this.
                    pass

    def nextnum(self):
        if not self.lots:
            return 1
        return max([l.number for l in self.lots]) + 1

    def first_lot_by_commodity(self, commodity):
        return [s for s in self if str(s.amount.commodity) == str(commodity)][0]

    def subtract(self, amount):
        lots = []
        subtracted = amount - amount
        while subtracted < amount:
            try:
                l = self.first_lot_by_commodity(amount.commodity)
            except IndexError:
                raise NotEnough(amount - subtracted)
            to_reduce = min([l.amount.strip_annotations(), amount - subtracted])
            if str(to_reduce) == str(l.amount.strip_annotations()):
                lots.append(l)
                self.lots.remove(l)
            else:
                l.amount -= to_reduce.number()
                new_amount = l.amount - l.amount + to_reduce.number()
                lots.append(Lot(l.number,
                                l.date,
                                new_amount,
                                l.account))
            subtracted += to_reduce
        return lots


def matches(string, options):
    for option in options:
        if fnmatch.fnmatch(string, option):
            return True
    return False


def main():
    journal, s = gui.load_journal_and_settings_for_gui()
    accts, unused_commodities = journal.accounts_and_last_commodity_for_account()

    saleacct = common.prompt_for_account(
        sys.stdin, sys.stdout,
        accts, "Which account was the sold commodity stored in?",
        s.get("last_sellstock_account", None)
    )
    assert saleacct, "Not an account: %s" % saleacct
    s["last_sellstock_account"] = saleacct

    commissionsaccount = common.prompt_for_account(
        sys.stdin, sys.stdout,
        accts, "Which account to account for commissions?",
        s.get("last_commissions_account", None)
    )
    assert commissionsaccount, "Not an account: %s" % commissionsaccount
    s["last_commissions_account"] = commissionsaccount

    gainslossesacct = common.prompt_for_account(
        sys.stdin, sys.stdout,
        accts, "Which account to credit gains and losses?",
        s.get("last_gainslosses_account",
              "Capital:Recognized gains and losses")
    )
    assert gainslossesacct, "Not an account: %s" % gainslossesacct
    s["last_gainslosses_account"] = gainslossesacct

    target_amount = common.prompt_for_amount(
        sys.stdin, sys.stdout,
        "How many units of what commodity?", ledger.Amount("$ 1")
    )
    target_sale_price = common.prompt_for_amount(
        sys.stdin, sys.stdout,
        "What is the sale price of the commodity?", ledger.Amount("$ 1")
    )
    commission = common.prompt_for_amount(
        sys.stdin, sys.stdout,
        "What was the commission of the trade?", ledger.Amount("$ 1")
    )

    all_lots = Lots()
    lots_text = subprocess.check_output([
        'ledger', 'bal',
        '--lots', '--lot-dates', '--lot-prices',
        '--date-format=%Y-%m-%d', '--sort=date',
        '--balance-format=++ %(account)\n%(amount)\n',
        saleacct
    ])
    all_lots.parse_ledger_bal(lots_text)

    print("=========== Read ===========")
    for l in all_lots:
        print(l)

    lots_produced = all_lots.subtract(target_amount)

    print("========= Computed =========")
    for l in lots_produced:
        print(l)

    print("=========== Left ===========")
    for l in all_lots:
        print(l)

    lines = []
    tpl = "%s {%s}%s @ %s"
    datetpl = ' [%s]'
    for l in lots_produced:
        m = -1 * l.amount
        if m.commodity.details.date:
            datetext = datetpl % m.commodity.details.date.strftime("%Y-%m-%d")
        else:
            datetext = ''
        lines.append((
            l.account,
            tpl % (m.strip_annotations(),
                   m.commodity.details.price,
                   datetext,
                   target_sale_price)
        ))
        diff = (l.price - target_sale_price) * l.amount
        lines.append((gainslossesacct, diff))
    totalsale = target_sale_price * sum(
        l.amount.number() for l in lots_produced
    )
    lines.append((saleacct, totalsale - commission))
    lines.append((commissionsaccount, commission))

    lines = journal.generate_record(
        "Sale of %s" % (target_amount),
        datetime.date.today(), None, "",
        lines,
    )
    print("========== Record ==========")
    print("\n".join(lines))
    save = common.yesno(
        sys.stdin, sys.stderr,
        "Hit ENTER or y to save it to the file, BACKSPACE or n to skip saving: "
    )
    if save:
        journal.add_text_to_file(lines)



#!/usr/bin/python3

import argparse


def get_common_argparser():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--file', dest='file', action='store',
                        help='specify path to ledger file to work with')
    parser.add_argument('--price-db', dest='pricedb', action='store',
                        help='specify path to ledger price database to work with')
    return parser
